#include "TheoraMovie.h"
#include <sys/time.h>
#include <iostream>

using namespace std;

// code largely taken from example_player.c 
// and all the audio and timing stuff removed

TheoraMovie::TheoraMovie() :
m_Data(NULL),
m_File(NULL)
{
}

TheoraMovie::~TheoraMovie()
{
	Clear();
}

bool TheoraMovie::Open(const string &filename)
{
	Clear();
	
	m_File=fopen(filename.c_str(),"rb");
    if(m_File==NULL)
	{
		fprintf(stderr,"Unable to open '%s' for playback.\n", filename.c_str());
		return false;
    } 
	
	// start up Ogg stream synchronization layer 
	ogg_sync_init(&oy);

	/* init supporting Theora structures needed in header parsing */
	theora_comment_init(&tc);
	theora_info_init(&ti);

	// Ogg file open; parse the headers 
	// Only interested in Theora streams 
	bool done=false;
 	while(!done)
	{
		if (!BufferData(m_File,&oy)) break; 

    	while(ogg_sync_pageout(&oy,&og)>0)
		{
			ogg_stream_state test;

			// is this a mandated initial header? If not, stop parsing 
			if(!ogg_page_bos(&og))
			{
				// don't leak the page; get it into the appropriate stream 
				ogg_stream_pagein(&to,&og);
				done=true;
				break;
			}

			ogg_stream_init(&test,ogg_page_serialno(&og));
			ogg_stream_pagein(&test,&og);
			ogg_stream_packetout(&test,&op);

			// identify the codec: try theora 
			if(!theora_p && theora_decode_header(&ti,&tc,&op)>=0)
			{
				// it is theora
				memcpy(&to,&test,sizeof(test));
				theora_p=1;
			}
			else
			{
				/* whatever it is, we don't care about it */
				ogg_stream_clear(&test);
      		}
		}
	}

  	// we're expecting more header packets. */
	while(theora_p && theora_p<3)
	{
		int ret;
		// look for further theora headers 
		while(theora_p && (theora_p<3) && (ret=ogg_stream_packetout(&to,&op)))
		{
			if(ret<0 || theora_decode_header(&ti,&tc,&op))
			{
				cerr<<"Error parsing Theora stream headers; corrupt stream?"<<endl;
				return false;
			}
			theora_p++;
			if(theora_p==3) break;
		}

		// The header pages/packets will arrive before anything else we
		// care about, or the stream is not obeying spec 
		if(ogg_sync_pageout(&oy,&og)>0)
		{
			ogg_stream_pagein(&to,&og); // demux into the appropriate stream 
		}
		else
		{
			if(BufferData(m_File,&oy)==0)
			{
				cerr<<"End of file while searching for codec headers."<<endl;
				return false;
			}
		}
	}

	// and now we have it all.  initialize decoders 
	if(theora_p)
	{
		theora_decode_init(&td,&ti);
	}
	else
	{
		Clear();
	}
	m_StartPos = ftell(m_File);
	return true;
}

void TheoraMovie::Clear()
{
	if (m_Data) delete[] m_Data;
	if (m_File) fclose(m_File);
	//theora_info_clear(&ti);
	//theora_comment_clear(&tc);
	theora_p=0;
}

void TheoraMovie::NextFrame()
{

	if(ogg_stream_packetout(&to,&op)>0)
	{
		theora_decode_packetin(&td,&op);
	}
	else
	{
		// no data yet for somebody.  Grab another page 
		BufferData(m_File,&oy);
		while(ogg_sync_pageout(&oy,&og)>0)
		{
			ogg_stream_pagein(&to,&og);
		}
	}
    
    GetVideoData();
	
    if(feof(m_File)) 
	{
		fseek(m_File,m_StartPos,SEEK_SET);
	}
}


int TheoraMovie::BufferData(FILE *in,ogg_sync_state *oy)
{
	char *buffer=ogg_sync_buffer(oy,4096);
	int bytes=fread(buffer,1,4096,in);
	ogg_sync_wrote(oy,bytes);
	return bytes;
}

void TheoraMovie::GetVideoData()
{
	yuv_buffer yuv;
	theora_decode_YUVout(&td,&yuv);

	if (m_Data) delete[] m_Data;
	m_Data=new char[yuv.y_width*yuv.y_height*3];
	int src=0;
	int uvsrc=0;
	int dst=0;

	int crop_offset=ti.offset_x+yuv.y_stride*ti.offset_y;
	int uvcrop_offset=(ti.offset_x/2)+yuv.uv_stride*(ti.offset_y/2);
	
	for (int r=0; r<yuv.y_height; r++)
	{
		src=crop_offset+yuv.y_stride*r;
		uvsrc=uvcrop_offset+yuv.uv_stride*(r/2);
		
		for (int x=0; x<yuv.y_width; x++)
		{
			// convert yuv->rgb
			float y=yuv.y[src+x];
			float u=yuv.u[uvsrc+x/2];
			float v=yuv.v[uvsrc+x/2];
			
			float b = 1.164f*(y-16.0f)+2.018f*(u-128.0f);
			float g = 1.164f*(y-16.0f)-0.813f*(v-128.0f)-0.391f*(u-128.0f);
			float r = 1.164f*(y-16.0f)+1.596f*(v-128.0f);
			
			// clamp
			if (r<0) r=0;
			if (g<0) g=0;
			if (b<0) b=0;
			if (r>255) r=255;
			if (g>255) g=255;
			if (b>255) b=255;			
			
			m_Data[dst++]=(unsigned char)r;
			m_Data[dst++]=(unsigned char)g;
			m_Data[dst++]=(unsigned char)b;
		}
	}
}
